from numba import types, njit, guvectorize, vectorize, prange, generated_jit
from numba.experimental import jitclass, structref
from numba import deferred_type, optional, objmode
from numba import void,b1,u1,u2,u4,u8,i1,i2,i4,i8,f4,f8,c8,c16
from numba.typed import List, Dict
from numba.core.types import (DictType, ListType, unicode_type, float64, NamedTuple, NamedUniTuple, UniTuple, Array, Tuple)
from numba.core.extending import (
    infer_getattr,
    lower_getattr_generic,
    lower_setattr_generic,
    overload_method,
    intrinsic,
    overload,
    box,
    unbox,
    NativeValue
)
from numba.core.datamodel import default_manager, models
from numba.core import types, cgutils
from numba.types import ListType
from numba.core.typing import signature

# from numba.core.extending import overload

from cre.core import TYPE_ALIASES, JITSTRUCTS, py_type_map, numba_type_map, numpy_type_map, register_global_default, lines_in_type_registry, add_to_type_registry, add_type_pickle
from cre.context import cre_context
from cre.caching import unique_hash_v, source_to_cache, import_from_cached, source_in_cache, get_cache_path
from cre.structref import gen_structref_code, define_structref
# from cre.context import cre_context
from cre.utils import (cast, struct_get_attr_offset, _obj_cast_codegen, 
                       _ptr_from_struct_codegen, CastFriendlyMixin, _obj_cast_codegen,
                       PrintElapse, _struct_get_data_ptr, _ptr_from_struct_incref,
                       deref_info_type, DEREF_TYPE_ATTR, DEREF_TYPE_LIST,
                       _ptr_to_data_ptr, _list_base_from_ptr)
from cre.obj import CREObjTypeClass, cre_obj_field_dict, CREObjModel, CREObjType, member_info_type, CREObjProxy

from numba.core.typeconv import Conversion
import operator
from numba.core.imputils import (lower_cast)
import cloudpickle
import numpy as np


SPECIAL_SPEC_ATTRIBUTES = ["inherit_from"]

class Fact(CREObjTypeClass):
    def __init__(self, name, fields, hash_code=None):
        if(hash_code is None):
            hash_code = unique_hash_v([name,fields])
        super(Fact, self).__init__(fields)
    
        # Sets the numba type name. Using hash_code is much shorter
        #  and thus more readable than using the fields dict.        
        self.name = f'{name}_{hash_code}'

    def __str__(self):
        if(hasattr(self, '_fact_name')):
            return f'{self._fact_name}_{self._hash_code[:10]}'
        elif(hasattr(self,"_specialization_name")):
            return self._specialization_name
        else:
            return "Fact"

    __repr__ = __str__

    def preprocess_fields(self, fields):
        return tuple((name, types.unliteral(typ)) for name, typ in fields)

    @property
    def field_dict_keys(self):
        if(not hasattr(self, "_field_dict_keys")):
            self._field_dict_keys = list(self.field_dict.keys())
        return self._field_dict_keys

    def get_attr_deref_info(self, attr):
        from cre.context import cre_context
        
        priv_var_name = f"_{attr}_deref_info"
        if(not hasattr(self, priv_var_name)):
            a_id = self.get_attr_a_id(attr)
            offset = self.get_attr_offset(attr)
            head_type = self.get_attr_type(attr)
            
            head_t_id = cre_context().get_t_id(_type=head_type)
            # deref_type = DEREF_TYPE_LIST if isinstance(head_type,types.ListType) else DEREF_TYPE_ATTR

            deref_info = np.zeros((),dtype=deref_info_type)
            deref_info['type'] = DEREF_TYPE_ATTR
            deref_info['a_id'] = a_id
            deref_info['t_id'] = head_t_id
            deref_info['offset'] = offset

            setattr(self, priv_var_name, deref_info)
        return getattr(self, priv_var_name)
        
    def get_attr_type(self, attr):
        attr_type = self.spec[attr]['type']
        if(isinstance(attr_type, ListType)):
            if(isinstance(attr_type.item_type,DeferredFactRefType)):
                attr_type = types.ListType(attr_type.item_type.get())
        if(isinstance(attr_type, DeferredFactRefType)):
            attr_type = attr_type.get()
        return attr_type

    def get_attr_offset(self,attr):
        return self._attr_offsets[self.get_attr_a_id(attr)]

    def get_attr_a_id(self, attr):
        return u1(self.field_dict_keys.index(attr))

    def get_attr_from_a_id(self, a_id):
        # print(self.name,a_id, self.field_dict_keys)
        return self.field_dict_keys[a_id]

    def __call__(self, *args, **kwargs):
        ''' If a fact_type is called with types return a signature 
            otherwise use it's ctor to return a new instance'''
        if len(args) > 0 and isinstance(args[0], types.Type):
            return signature(self, *args)

        # Ensure that any lists are converted to typed lists
        args = [List(x) if(isinstance(x,list)) else x for x in args]
        kwargs = {k: (List(v) if(isinstance(v,list)) else v) for k,v in kwargs.items()}
        return self._ctor[0](*args, **kwargs)

    def filter_spec(self, flags):
        ''' Returns a filtered version of this Fact's spec which picks
            out only the subset of attributes in the spec that satisfy
            a set of flags like [('visible', 'few_valued'), ('unique_id')]
            where the list of lists or tuples of flags represents a disjunctive 
            normal statement i.e. ('visible' and 'few_valued') or 'unique_id'.
        '''

        # Ensure that flags is list of lists
        if(not isinstance(flags,(list,tuple))): flags = [flags]
        if(len(flags) == 0): return self.clean_spec
        flags = list(flags)
        if(not isinstance(flags[0], (list, tuple))): flags = [flags]

        from cre.attr_filter import attr_filter_registry
        filtered_spec = {}
        for and_flags in flags:
            flag_0 = and_flags[0]
            negated_0 = flag_0[0] == "~"
            flag_0 = flag_0[1:] if(negated_0) else flag_0
            if(flag_0 not in attr_filter_registry):
                raise ValueError(f"No filter registered under flag {flag_0!r}")
            conj_filtered_spec = {k:v for k,v in attr_filter_registry[flag_0].get_attrs(self, negated_0)}
            for i in range(1,len(and_flags)):
                flag = and_flags[i]
                negated = flag[0] == "~"
                flag = flag[1:] if negated else flag
                if(flag not in attr_filter_registry):
                    raise ValueError(f"No filter registered under flag {flag!r}")
                _fs = attr_filter_registry[flag].get_attrs(self, negated)
                conj_filtered_spec = {k : v for k,v in _fs if k in conj_filtered_spec}
            filtered_spec = {**filtered_spec, **conj_filtered_spec}
        return filtered_spec

    @property
    def clean_spec(self):
        if(not hasattr(self,'_clean_spec')):
            self._clean_spec = _clean_spec(self.spec)
        return self._clean_spec

    def __getstate__(self):
        d = self.__dict__.copy()
        if('_clean_spec' in d): del d['_clean_spec']

        # NOTE: While ops still use py_classes we need clean out 'conversions' from the spec
        if('spec' in d):

            d_spec = d['spec'].copy()
            
            for attr, attr_spec in d_spec.items():
                if('conversions' in attr_spec):
                    attr_spec_copy = attr_spec.copy()
                    attr_spec_copy['conversions'] = tuple([(attr_spec['type'], attr, typ) for typ in attr_spec_copy['conversions']])
                    d_spec[attr] = attr_spec_copy
            d['spec'] = d_spec
            # print(d)
        if('_code' in d): del d['_code']
                    
        return d

    def __setstate__(self, d):
        self.__dict__ = d





@overload_method(Fact, '__str__')
def overload_str(self):
    def impl(self):
        with objmode(s='unicode_type'):
            s = str(self)
        return s
    return impl


@overload_method(Fact, '__repr__')
def overload_repr(self):
    def impl(self):
        with objmode(s='unicode_type'):
            s = repr(self)
        return s
    return impl

@generated_jit(cache=True)
def call_untyped_fact(self, **kwargs):

    typs = tuple(kwargs.values())
    return_sig = False
    if(isinstance(typs[0],types.TypeRef)):
        typs = tuple(x.instance_type for x in typs)
        return_sig = True    
    fact_type = self.specialize(**{k:v for k,v in zip(kwargs,typs)})

    if(return_sig):
        def impl(self,**kwargs):
            return fact_type
    else:
        def impl(self,**kwargs):
            return fact_type(**kwargs)

    return impl

from numba.core.typing.typeof import typeof
class UntypedFact(Fact):
    specializations = {}
    def __init__(self, name, fields, hash_code=None):
        super(UntypedFact, self).__init__(name, fields, hash_code)
    def __call__(self, *args, **kwargs):
        if(len(kwargs) == 0):
            return signature(self,*args)
        else:

            if(isinstance(list(kwargs.values())[0],types.Type)):
                fact_type = self.specialize(**kwargs)
                return fact_type#call_untyped_fact(self, **kwargs)
            else:
                fact_type = self.specialize(**{k:typeof(v) for k,v in kwargs.items()})
                ctor = fact_type._ctor[0]
                return ctor(**kwargs)


    def specialize(self, **kwargs):
        typs = tuple(kwargs.values())
        self_field_dict = {k:v for (k,v) in zip(kwargs.keys(), typs)}
        spec = {**self_field_dict, "inherit_from" : self}
        fact_type = define_fact(self._fact_name, spec)
        return fact_type



    # @property
    # def key(self):
    #     """
    #     A property used for __eq__, __ne__ and __hash__.  Can be overridden
    #     in subclasses.
    #     """
    #     # if(hasattr(self,'name'))
    #     return self.name
    #     # return (self._fact_name, getattr(self,"_hash_code",None), getattr())

    

    # def __getstate__(self):
    #     state = self.__dict__.copy()
    #     if(hasattr(state,'spec')): del state['spec']
    #     if(hasattr(state,'fact_ctor')): del state['fact_ctor']
    #     return state

    # def __setstate__(self,state):
    #     self.__dict__.update(state)
    #     state = self.__dict__.copy()
        
    #     return state







###### Fact Specification Preprocessing #######
class DeferredFactRefType():
    '''A placeholder type for when referencing a fact type that
        is not defined yet. Note: Sort of mimics deferred_type but
         doesn't subclass because would use special numba pickling,
         which I haven't quite figured out.
    '''
    def __init__(self,typ):
        self._fact_name = typ._fact_name if isinstance(typ, types.StructRef) else typ 
        super(DeferredFactRefType,self).__init__()
    def __eq__(self,other):
        return isinstance(other,DeferredFactRefType) \
               and self._fact_name == other._fact_name
    def __str__(self):
        return f"DeferredFactRefType[{self._fact_name}]"

    def __repr__(self):
        return f"DeferredFactRefType({self._fact_name!r})"

    @property
    def instance_type(self):
        return getattr(self,'_definition', None)

    def define(self,x):
        self._definition = x

    def get(self):
        if(not hasattr(self,"_definition")): raise TypeError(f"Attempt to use {str(self)} without definition")
        return self._definition

    def __hash__(self):
        return hash(self._fact_name)

    def __setstate__(self,state):
        self._fact_name = state[0]

    def __getstate__(self):
        return (self._fact_name,)#({'_fact_name' : self._fact_name})



def _standardize_type(typ, context, name='', attr=''):
    '''Takes in a string or type and returns the standardized type'''
    if(isinstance(typ, type)):
        typ = typ.__name__
    if(isinstance(typ,str)):
        typ_str = typ
        is_list = typ_str.lower().startswith("list")
        if(is_list): typ_str = typ_str.split("(")[1][:-1]


        is_deferred = False
        if(typ_str.lower() in TYPE_ALIASES): 
            typ = numba_type_map[TYPE_ALIASES[typ_str.lower()]]
        # elif(typ_str == name):
        #     typ = context.get_deferred_type(name)# DeferredFactRefType(name)
        elif(typ_str in context.name_to_type):
            typ = context.name_to_type[typ_str]
        else:
            typ = context.get_deferred_type(typ_str)
            is_deferred = True
            # raise TypeError(f"Attribute type {typ_str!r} not recognized in spec" + 
            #     f" for attribute definition {attr!r}." if attr else ".")

        if(is_list): typ = ListType(typ)

    if(hasattr(typ, "_fact_type")): typ = typ._fact_type
    return typ


def _merge_spec_inheritance(spec : dict, context):
    '''Expands a spec with attributes from its 'inherit_from' type'''
    if("inherit_from" not in spec): return spec, None
    inherit_from = spec["inherit_from"]

    unified_attrs = []
    if(isinstance(inherit_from, dict)):
        unified_attrs = inherit_from.get('unified_attrs')
        inherit_from = inherit_from['type']

    if(isinstance(inherit_from, str)):
        temp = inherit_from
        inherit_from = context.name_to_type[inherit_from]
    if(not isinstance(inherit_from,types.StructRef)):
        inherit_from = context.name_to_type[inherit_from._fact_name]
        
        
    if(not hasattr(inherit_from, 'spec')):
        raise ValueError(f"Invalid inherit_from : {inherit_from}")

    inherit_spec = inherit_from.spec

    _intersect = set(inherit_spec.keys()).intersection(set(spec.keys()))
    for k in _intersect:
        if(k in unified_attrs): continue
        if(spec[k]['type'] != inherit_spec[k]['type']): 
            raise TypeError(f"Attribute type {k}:{spec[k]['type']} does not" +
                            f"match inherited attribute {k}:{inherit_spec[k]['type']}")
    del spec['inherit_from']
    return {**inherit_spec, **spec}, inherit_from

def _standardize_conversions(conversions, attr_type, context):
    from cre.func import UntypedCREFunc
    assert isinstance(conversions, dict), f"'conversions' expecting dict : type -> conversion_op, not {type(conversions)}."
    stand_conv = {}
    for conv_type, conv_op in conversions.items():
        conv_type = _standardize_type(conv_type, context)
        print(attr_type)
        if(isinstance(conv_op, UntypedCREFunc)): conv_op = conv_op(attr_type)
        assert conv_op.return_type == conv_type, f"{conv_op} does not return conversion type {conv_type}."
        stand_conv[conv_type] = conv_op
    return stand_conv

    # attr_spec['conversions'] = {_standardize_type(k, context):v for }

def _standardize_spec(spec : dict, context, name=''):
    '''Takes in a spec and puts it in standard form'''

    out = {}
    for attr, attr_spec in spec.items():
        if(attr in SPECIAL_SPEC_ATTRIBUTES): out[attr] = attr_spec; continue;

        if(isinstance(attr_spec, dict) and not "type" in attr_spec):
            raise AttributeError("Attribute specifications must have 'type' property, got %s." % v)
        
        typ, attr_spec = (attr_spec['type'], attr_spec) if isinstance(attr_spec, dict) else (attr_spec, {})
        typ = _standardize_type(typ, context, name, attr)

        if('conversions' in attr_spec): 
            attr_spec['conversions'] = _standardize_conversions(attr_spec['conversions'], typ, context)

        out[attr] = {"type": typ,**{k:v for k,v in attr_spec.items() if k != "type"}}
    return out

def _clean_spec(spec : dict):
    '''Replaces any defferred types in a spec with their definitions'''
    new_spec = {}
    for attr, attr_spec in spec.items():
        attr_t = attr_spec['type']
        attr_t = attr_t.instance_type if (isinstance(attr_t, (types.TypeRef,DeferredFactRefType))) else attr_t

        # Handle List case
        if(isinstance(attr_t, types.ListType)):
            item_t = attr_t.item_type
            item_t = item_t.instance_type if (isinstance(item_t, (types.TypeRef,DeferredFactRefType))) else item_t
            attr_t = types.ListType(item_t)
        new_spec[attr] = {**attr_spec, 'type': attr_t}
    return new_spec








###### Fact Definition #######
class FactProxy(CREObjProxy):
    # '''Essentially the same as numba.experimental.structref.StructRefProxy 0.51.2
    #     except that __new__ is not defined to statically define the constructor.
    # '''
    # __slots__ = ('_fact_type', '_meminfo')

    @classmethod
    def _numba_box_(cls, mi):
        """Called by boxing logic, the conversion of Numba internal
        representation into a PyObject.

        Parameters
        ----------
        mi :
            a wrapped MemInfoPointer.

        Returns
        -------
        instance :
             a FactProxy instance.
        """
        inst = super(FactProxy,cls)._numba_box_(BaseFact,mi)
        return inst

    @classmethod
    def _numba_box_no_recover_(cls, ty, mi):
        '''Same as StructRefProxy'''
        instance = ty.__new__(cls)
        instance._type = ty
        instance._meminfo = mi
        return instance


    @property
    def _numba_type_(self):
        """Returns the Numba type instance for this structref instance.

        Subclasses should NOT override.
        """
        return self._fact_type

    # def __eq__(self, other):
    #     from cre.dynamic_exec import fact_eq
    #     if(isinstance(other, FactProxy)):
    #         return fact_eq(self,other)
    #     return False

    def restore(self,context=None):
        context = cre_context(context)
        # if(context.tt_id):

    # def __hash__(self):
    #     from cre.dynamic_exec import fact_hash
    #     return fact_hash(self)

    def get_ptr(self):
        return fact_to_ptr(self)

    def get_ptr_incref(self):
        return fact_to_ptr_incref(self)

    def _gen_val_var_possibilities(self, self_var):
        for attr, config in self._fact_type.clean_spec.items():
            typ = config['type']
            # with PrintElapse("getattr_var"):
            val = getattr(self,attr)
            # with PrintElapse("getattr_var"):
            attr_var = getattr(self_var, attr)
            if(isinstance(val, List)):
                for i in range(len(val)):
                    item_var = attr_var[i]
                    item_val = val[i]
                    yield (item_val, item_var)
            else:
                yield (val, attr_var)
            # else:
                # Primitive case
                # one_lit_conds.append(attr_var==val)

    def as_conditions(self, fact_ptr_to_var_map=None, keep_null=True, add_implicit_neighbor_self_refs=True, neigh_count = 0):
        from cre.default_funcs import Equals
        from cre.utils import as_typed_list
        from cre.dynamic_exec import var_eq
        from cre.var import Var



        self_ptr = self.get_ptr()

        if(fact_ptr_to_var_map is None):
             fact_ptr_to_var_map = {self_ptr : Var(self._fact_type, "X")}
            

        self_var = fact_ptr_to_var_map[self_ptr]
        one_lit_conds = []

        

        
        # with PrintElapse("CONSTRUCTS"):       
        # for attr, config in self._fact_type.spec.items():
        for attr_val, attr_var in self._gen_val_var_possibilities(self_var):
            if(isinstance(attr_val, FactProxy)):
                # Fact case
                attr_val_fact_ptr = attr_val.get_ptr()
                if(attr_val_fact_ptr not in fact_ptr_to_var_map):
                    if(add_implicit_neighbor_self_refs):
                        nbr_var = Var(attr_val._fact_type, f"Nbr{neigh_count}")
                        fact_ptr_to_var_map[attr_val_fact_ptr] = nbr_var
                        neigh_count += 1
                    else:
                        continue


                val_var = fact_ptr_to_var_map[attr_val_fact_ptr]
                #   FIXME: use cre_obj.__eq__()
                
                    # str(attr_var) == str(val_var)
                # print("<<", str(attr_var), str(val_var), var_eq(attr_var, val_var))
                if(add_implicit_neighbor_self_refs and str(attr_var) == str(val_var)):
                    # for case like x.next == x.next, try make conditions like x == x.next.prev
                    # with PrintElapse("LOOP"):
                        #     list(attr_val._gen_val_var_possibilities(attr_var))
                    for attr_val2, attr_var2 in attr_val._gen_val_var_possibilities(attr_var):

                        if(isinstance(attr_val2, FactProxy) and 
                            attr_val2.get_ptr() == self_ptr):
                            one_lit_conds.append(self_var==attr_var2)

                else:
                    # with PrintElapse("NEW LIT"):
                    one_lit_conds.append(attr_var==fact_ptr_to_var_map[attr_val_fact_ptr])
                
            else:
                # Note: Making literals with primitives is slow w/ current Op 
                #  implmenentation since it compiles compositions. 

                # Primitive case
                if(not keep_null and attr_val is None): continue
                one_lit_conds.append(attr_var==attr_val)

        # with PrintElapse("ANDS"):    
        # print(fact_ptr_to_var_map)
        _vars = list({v.get_ptr():v for v in fact_ptr_to_var_map.values()}.values())
        # print(_vars)   
        conds = _vars[0]
        for i in range(1, len(_vars)):
            conds = conds & _vars[i]

        for c in one_lit_conds:
            conds = conds & c

        return conds

    def isa(self, typ):
        return isa(self,typ)

    def asa(self, typ):
        if(not isa(self,typ)):
            raise TypeError(f"Cannot cast fact '{str(self)}' to '{str(typ)}.'")
        return super().asa(typ)
    # #     instance = typ.__new__(CREObjProxy)
    # #     instance._type = typ
    # #     instance._meminfo = self._meminfo
    # #     return instance
        # return asa(self,typ)

    def __repr__(self):
        return str(self)


    def resolve_deref(self, derefs):
        if(not isinstance(derefs, np.ndarray)):
            derefs = derefs.deref_infos
        ctx = cre_context()
        out_type = ctx.get_type(t_id=derefs[-1]['t_id'])
        return resolve_deref(self, derefs, out_type)

    # def __eq__(self, other):
    #     from cre.dynamic_exec import fact_eq
    #     return fact_eq(self.asa(CREObjType), other.asa(CREObjType))

    # def __hash__(self):
    #     from cre.dynamic_exec import fact_hash
    #     return fact_hash(self.asa(CREObjType))


def gen_fact_import_str(t):
    return f"from cre_cache.{t._fact_name}._{t._hash_code} import {t._fact_name}"

def gen_inherit_import_str(t):
    return f"from cre_cache.{t._fact_name}._{t._hash_code} import {t._fact_name} as parent_type, inheritance_bytes as parent_inh_bytes"

def _gen_getter_jit(f_typ,typ,attr):
    if(isinstance(typ,(Fact,DeferredFactRefType))):
        return \
f'''@njit(cache=True)
def get_{attr}_as_ptr(self):
    return get_fact_attr_ptr(self, '{attr}')

@njit(cache=True)
def get_{attr}(self):
    return self.{attr}
'''
    else:
        return \
f'''@njit(cache=True)
def get_{attr}(self):
    return self.{attr}
'''

def _gen_setter_jit(f_typ, attr, a_id):
    return f'''@njit(types.void({f_typ},field_list[{a_id}][1]), cache=True)
def set_{attr}(self, val):
    fact_mutability_protected_setattr(self,'{attr}',val)
'''

def _gen_props(attr):
    return f'''    {attr} = property(get_{attr}, set_{attr})'''

def get_type_default(t):
    if(isinstance(t,(bool,types.Boolean))):
        return False
    elif(isinstance(t,(str,types.UnicodeType))):
        return ""
    elif(isinstance(t,(float,types.Float))):
        return 0.0
    elif(isinstance(t,(int,types.Integer))):
        return 0
    else:
        return None



fact_types = (Fact, DeferredFactRefType)

#### Resolving Byte Offsets of Struct Members ####

def get_offsets_from_member_types(fields):
    # from cre.fact import fact_types, FactModel, BaseFact
    if(isinstance(fields, dict)): fields = [(k,v) for k,v in fields.items()]
    #Replace fact references with BaseFact
    # fact_types = (types.StructRef, DeferredFactRefType)
    fields = [(a,BaseFact if isinstance(t,fact_types) else t) for a,t in fields]

    class TempTypeTemplate(types.StructRef):
        pass

    default_manager.register(TempTypeTemplate, CREObjModel)

    TempType = TempTypeTemplate(fields)

    return [struct_get_attr_offset(TempType,attr) for attr, _ in fields]


def repr_fact_attr(inst):
    if(inst is None): return 'None'

    inst_type = type(inst)
    # cre_context().get
    # print(isinstance(inst, Fact), str(inst_type))
    if(hasattr(inst_type,"_fact_type") and
        hasattr(inst_type._fact_type, "_specialization_name")):
        return str(inst)

    ptr = inst.get_ptr()
    if(ptr != 0):
        return f"<{inst_type._fact_name} at {hex(ptr)}>"
    else:
        return 'None'

def repr_list_attr(val,dtype_name=None):
    # if(isinstance(val,Fact)):
    # ptr = get_ptr(val)
    if(val is not None):
        if(dtype_name is not None):
            return f'List([{", ".join([f"<{dtype_name} at {hex(fact_to_ptr(x))}>" for x in val])}])'
        else:
            return f'List([{", ".join([repr(x) for x in val])}])'
    else:
        return 'None'



def gen_repr_attr_code(a,t):
    '''Helper function for generating code for the repr/str of the fact'''
    if(isinstance(t,fact_types)):
        return f'{a}={{repr_fact_attr(self.{a})}}'
        # return f'{a}=<{t._fact_name} at {{hex({typ_name}_get_{a}_as_ptr(self))}}>'
    elif(isinstance(t,ListType)):
        # TODO : might want to print lists like reference where just the address is printed
        # if():
        s = ", " + f'"{t.dtype._fact_name}"' if isinstance(t.dtype,fact_types) else ""
        return f'{a}={{repr_list_attr(self.{a}{s})}}'
            # s = f'f"<{t.dtype._fact_name} at {{hex(fact_to_ptr(x))}}>"'
            # return f'{a}={{"List([" + ", ".join([{s} for x in self.{a}]) + "])" if self.{a} is not None else "None"}}'
        # else:
            # return f'{a}=List([{{", ".join([repr(x) for x in self.{a}])}}])'
    else:
        return f'{a}={{repr(self.{a})}}' 

# def gen_assign_str(a,t):
#     # if(isinstance(t,fact_types)):
#         # s = f"_ptr_from_struct_incref({a}) if ({a} is not None) else 0"
#     # elif(isinstance(t,ListType) and isinstance(t.dtype,fact_types)):
#     #     # s = f"{a}_c = _cast_list(base_list_type,{a})\n    "
#     #     # s += f"st.{a} = {a}_c" 
#     #     # s = f"st.{a} = _cast_list(base_list_type,{a})" 
#     #     s = f"st.{a} = {a}"
#     # else:
#     s = f"{a}"

#     return f"st.{a} = " + s

@njit(u1[::1](u8),cache=True)
def uint_to_inheritance_bytes(n):
    buffer = np.empty((8,), dtype=np.uint8)
    i = 0
    while(n != 0):
        buffer[8-(i+1)] = n & 0xFF
        i += 1
        n = n >> 8
    return buffer[-i:]

from cre.obj import member_info_type
from cre.utils import _sizeof_type, _load_ptr



def _prep_field(attr, t, imports_set):
    from cre.tuple_fact import TupleFactClass
    # elif(isinstance(t,types.ListType) and isinstance(t.dtype,types.StructRef)):
    #     imports_set.add(f"{gen_fact_import_str(t.dtype)}")

    # upcast any facts to BaseFact since references to undefined fact types not supported
    if(isinstance(t,fact_types)):
        if(isinstance(t,Fact) and not isinstance(t,TupleFactClass)):
            imports_set.add(f"{gen_fact_import_str(t)}")
        return attr, BaseFact
    elif(isinstance(t,ListType)):
        if(isinstance(t.dtype,fact_types)):
            _, dtype = _prep_field(attr, t.dtype, imports_set)
            # if(isinstance(dtype, types.Optional)): dtype = dtype.type

            return (attr, ListType(dtype))
        return attr, t
    else:
        return attr, t


def _prep_fields_populate_imports(fields, inherit_from=None):
    imports_set = set()
    if(inherit_from is not None):
        imports_set.add(f"{gen_inherit_import_str(inherit_from)}")
    fields = [_prep_field(attr,t,imports_set) for attr, t in fields]
        
    return fields, "\n".join(list(imports_set))

def gen_fact_src(typ, fields, t_id, inherit_from=None, specialization_name=None, is_untyped=False, hash_code="", ind='    '):
    '''Generate the source code for a new fact '''
    fields, fact_imports = _prep_fields_populate_imports(fields, inherit_from)

    _base_fact_field_dict = {**base_fact_field_dict}
    all_fields = [(k,v) for k,v in _base_fact_field_dict.items()] + fields    # all_fields = [(k,v) for (k,v) in all_fields]

    # all_fields = base_fact_fields+fields
    properties = "\n".join([_gen_props(attr) for attr,t in all_fields])
    getter_jits = "\n".join([_gen_getter_jit(typ,t,attr) for attr,t in all_fields])
    setter_jits = "\n".join([_gen_setter_jit(typ,attr,a_id) for a_id, (attr,t) in enumerate(all_fields)])
    # field_list = ",".join(["'%s'"%attr for attr,t in fields])

    param_defaults_seq = ",".join([f"{attr}={get_type_default(t)!r}" for attr,t in fields])
    param_seq = ",".join([f"{attr}" for attr,t in fields])
    attr_tup = tuple([attr for attr,t in fields])


    as_dict_setters = []
    for attr,t in fields:
        print(attr, t)
        if(isinstance(t, Fact)):
            s = f"'{attr}':getattr(get_{attr}(self), key_attr, None)"
        elif(isinstance(t, ListType)):
            if(isinstance(t.item_type, Fact)):
                s = f"'{attr}':[getattr(x, key_attr, None) for x in get_{attr}(self)]"
            else:
                s = f"'{attr}':list(get_{attr}(self))"
        else:
            s = f"'{attr}':get_{attr}(self)"

        as_dict_setters.append(s)
    as_dict_setters = ", ".join(as_dict_setters)

    # base_list = ",".join([f"'{k}'" for k,v in _base_fact_fields])
    # base_type_list = ",".join([str(v) for k,v in _base_fact_fields])

    # fact_types = (types.StructRef, DeferredFactRefType)
    
    # field_type_list = ",".join([repr_type(v) for k,v in fields])


    # assign_str = lambda a,t: f"st.{a} = " + (f"_ptr_from_struct_incref({a}) if ({a} is not None) else 0" \
    #                         if isinstance(t,fact_types) else f"{a}")
    init_fields = f'\n{ind}'.join([f"fact_lower_setattr(st,'{k}',{k})" for k,v in fields])

    str_temp = ", ".join([gen_repr_attr_code(k,v) for k,v in fields])

    #TODO get rid of this

    attr_offsets = get_offsets_from_member_types(all_fields)

# The source code template for a user defined fact. Written to the
#  system cache so it can be its own module. Doing so helps njit(cache=True)
#  work when using user defined facts.
    code = \
f'''
import numpy as np
from numba.core import types
from numba import njit, literally, literal_unroll
from numba.core.types import *
from numba.core.types import unicode_type, ListType, UniTuple, Tuple
from numba.experimental import structref
from numba.experimental.structref import new#, define_boxing
from numba.core.extending import overload, overload_method, lower_cast, type_callable
from numba.core.imputils import numba_typeref_ctor
from cre.fact_intrinsics import define_boxing, get_fact_attr_ptr, _register_fact_structref, fact_mutability_protected_setattr, fact_lower_setattr, _fact_get_chr_mbrs_infos
from cre.fact import repr_list_attr, repr_fact_attr, FactProxy, Fact, UntypedFact{", BaseFact, base_list_type, fact_to_ptr, get_inheritance_bytes_len_ptr" if typ != "BaseFact" else ""}, uint_to_inheritance_bytes
from cre.utils import cast, ptr_t, _get_member_offset,  _load_ptr, _obj_cast_codegen, encode_idrec
import cloudpickle
from cre.obj import member_info_type, set_chr_mbrs
{fact_imports}



attr_offsets = np.array({attr_offsets!r},dtype=np.int16)
inheritance_bytes = tuple({"list(parent_inh_bytes) + [u1(0)] + " if inherit_from else ""}list(uint_to_inheritance_bytes({t_id}))) 
num_inh_bytes = len(inheritance_bytes)
hash_code = '{hash_code}'

@_register_fact_structref
class {typ}Class({"UntypedFact" if is_untyped else "Fact"}):
    def __init__(self, fields, hash_code=None):
        super().__init__('{typ}', fields, hash_code)
        self._fact_name = '{typ}'
        self.t_id = {t_id}
        self._attr_offsets = attr_offsets
        self._hash_code = '{hash_code}'
        {f'self._specialization_name = "{specialization_name}"' if(specialization_name is not None) else ''}

    

field_list = cloudpickle.loads({cloudpickle.dumps(all_fields)})
{typ} = fact_type = {typ}Class(field_list, hash_code)
{typ}_w_mbr_infos = {typ}Class(field_list+
[("chr_mbrs_infos", UniTuple(member_info_type,{len(fields)})),
 ("num_inh_bytes", u1),
 ("inh_bytes", UniTuple(u1, num_inh_bytes))])


{(f"""{typ}.parent_type = parent_type
pt = parent_type
while(pt is not None):
    @lower_cast({typ}, pt)
    def upcast(context, builder, fromty, toty, val):
        return _obj_cast_codegen(context, builder, val, fromty, toty,incref=False)                        
    pt = getattr(pt, 'parent_type', None)
""") if inherit_from is not None else ""
}

@njit(cache=True)
def get_chr_mbrs_infos():
    st = new({typ})
    return _fact_get_chr_mbrs_infos(st)

chr_mbrs_infos = get_chr_mbrs_infos()
{typ}Class._chr_mbrs_infos = chr_mbrs_infos

{(f"""#locals={{'inheritance_bytes':Tuple((u1,num_inh_bytes))}}
@njit(u1(BaseFact), cache=True)
def isa_{typ}(fact):
    l, p = get_inheritance_bytes_len_ptr(fact)
    if(l >= num_inh_bytes):
        for i,b in enumerate(literal_unroll(inheritance_bytes)):
            f_b = _load_ptr(u1, p+i)
            if(b != f_b):
                return False
        return True
    else:
        return False
""") if typ != "BaseFact" else (
f"""@njit(u1(BaseFact), cache=True)
def isa_{typ}(fact):
    return True
"""
)
}
{typ}Class._isa = isa_{typ}

@njit(cache=True)
def ctor({param_defaults_seq}):
    st = new({typ}_w_mbr_infos)
    fact_lower_setattr(st,'idrec',encode_idrec({t_id},0,u1(-1)))
    fact_lower_setattr(st,'hash_val',0)
    set_chr_mbrs(st, {attr_tup!r})
    fact_lower_setattr(st,'num_inh_bytes', num_inh_bytes)
    fact_lower_setattr(st,'inh_bytes', inheritance_bytes)
    {init_fields}
    return cast(st, {typ})

# Put in a tuple so it doesn't get wrapped in a method
{typ}Class._ctor = (ctor,)

{getter_jits}

{setter_jits}
        
class {typ}Proxy(FactProxy):
    __numba_ctor = ctor
    _fact_type = {typ}
    _fact_type_class = {typ}Class
    _fact_name = '{typ}'
    {f'_specialization_name = "{specialization_name}"' if(specialization_name is not None) else ''}
    t_id = {t_id}
    _attr_offsets = attr_offsets
    _chr_mbrs_infos = chr_mbrs_infos
    _isa = isa_{typ}
    _code = {typ}._code
    _hash_code = '{hash_code}'

    def __new__(cls, *args,**kwargs):
        return ctor(*args,**kwargs)

    def __repr__(self):
        return f'{typ}({str_temp})'

    def __str__(self):
        return self.__repr__()

    def as_dict(self, key_attr='idrec'):
        return {{'type' : '{typ}', {as_dict_setters}}}


{properties}

# Overload '{typ}' as its own constructor
@type_callable({typ})
def ssp_call(context):
    {"raise NotImplementedError('NotImplementedError: Cannot initialize UntypedFact in jitted context.')" if is_untyped else ""}
    # Note to self this requires *args, see https://github.com/numba/numba/issues/7973
    def typer({param_seq}):    
        {f'return {typ}.specialize({param_seq})' if is_untyped else f'return {typ}'}
    return typer

@overload(numba_typeref_ctor)
def overload_{typ}(self, {param_defaults_seq}):
    if(self.instance_type is not {typ}): return
    def impl(self, {param_defaults_seq}):
        return ctor({param_seq})
    return impl

{typ}Class._fact_type = {typ}
{typ}Class._fact_proxy = {typ}Proxy
{typ}Class._proxy_class = {typ}Proxy

{typ}._fact_type_class = {typ}Class
{typ}._fact_proxy = {typ}Proxy
{typ}._proxy_class = {typ}Proxy


define_boxing({typ}Class,{typ}Proxy)



{(f"""from cre.var import VarType, var_ctor
# @njit(VarType(unicode_type), cache=True)
# def as_var(alias):
#     return var_ctor({typ}, {t_id}, alias)
# {typ}._as_var = as_var
""") if typ != "BaseFact" else ""
}


'''
    return code



def _fact_from_fields(name, fields, inherit_from=None, specialization_name=None, is_untyped=False, return_proxy=False, return_type_class=False):
    # context = cre_context(context)
    

    hash_code = unique_hash_v([name,fields])
    if(not source_in_cache(name,hash_code)):
        # Possible for other types to be defined while running the Fact source
        #  so preregister the t_id then add the pickle later.
        t_id = add_to_type_registry(name, hash_code)
        source = gen_fact_src(name, fields, t_id, inherit_from, specialization_name, is_untyped, hash_code)
        source_to_cache(name, hash_code, source)
        
        fact_type = tuple(import_from_cached(name, hash_code, [name]).values())[0]
        add_type_pickle(fact_type, t_id)
        
    to_get = [name]
    if(return_proxy): to_get.append(name+"Proxy")
    if(return_type_class): to_get.append(name+"Class")
        
    out = tuple(import_from_cached(name, hash_code, to_get).values())
    for x in out: x._hash_code = hash_code
        
    return tuple(out) if len(out) > 1 else out[0]

def _fact_from_spec(name, spec, inherit_from=None, specialization_name=None, return_proxy=False, return_type_class=False):
    # assert parent_fact_type
    is_untyped = bool(spec is None)
    fields = [(k,v['type']) for k, v in spec.items()] if spec else {}
    return _fact_from_fields(name, fields,
             inherit_from=inherit_from, specialization_name=specialization_name,
            is_untyped=is_untyped, return_proxy=return_proxy,
            return_type_class=return_type_class)

def _spec_eq(spec_a, spec_b):
    # print(list(spec_a.keys()), list(spec_b.keys()))
    for attr_a, attr_b in zip(spec_a, spec_b):
        if(attr_a != attr_b): 
            # print(attr_a, "!=", attr_b)
            return False
        typ_a, typ_b = spec_a[attr_a]['type'], spec_b[attr_a]['type']

        typ_strs = []
        for typ in [typ_a, typ_b]:
            if(isinstance(typ, ListType)):
                if(isinstance(typ.item_type,(DeferredFactRefType, Fact))):
                    item_str = typ.item_type._fact_name
                else:
                    item_str = str(typ.item_type)
                typ_strs.append(f"List({item_str})")

            elif(isinstance(typ, (DeferredFactRefType, Fact))):
                typ_strs.append(typ._fact_name)
            else:
                typ_strs.append(str(typ))

        if(typ_strs[0] != typ_strs[1]):
            # print(typ_strs[0], "!=", typ_strs[1])
            return False
    return True


def define_fact(name : str, spec : dict = None, context=None, return_proxy=False, return_type_class=False, allow_redef=False):
    '''Defines a new fact.'''

    from cre.context import cre_context
    context = cre_context(context)
    # print("DEFINE", name, context.name)
    specialization_name = name
    if(spec is not None):
        spec = _standardize_spec(spec,context,name)
        spec, inherit_from = _merge_spec_inheritance(spec, context)
    
        if(inherit_from is not None and 
            inherit_from._fact_name == name):
            # Specialize UntypedFact case
            fact_type = context.name_to_type[name]
            typ_assigments = ", ".join([f"{k}={str(v['type'])}" for k,v in spec.items() if k not in base_fact_field_dict])
            specialization_name = f"{name}({typ_assigments})"
    else:
        inherit_from = None


    if(specialization_name in context.name_to_type):
        assert _spec_eq(context.name_to_type[specialization_name].spec, spec), \
        f"Redefinition of fact '{specialization_name}' in context '{context.name}' not permitted"
        # print("SPECIALIZATION NAME:", specialization_name)
        fact_type = context.name_to_type[specialization_name]
    else:

        fact_type = _fact_from_spec(name, spec, inherit_from=inherit_from, 
            specialization_name= (specialization_name if specialization_name != name else None),
            return_proxy=False, return_type_class=False)
        dt = context.get_deferred_type(name)
        dt.define(fact_type)
        # context._assert_flags(name, spec)
        context._register_fact_type(specialization_name, fact_type, inherit_from=inherit_from)
    _spec = spec if(spec is not None) else {}
    # _spec = _undeffer_spec(_spec)
    # print({_id : str(config['type']) for _id, config in _spec.items()})
    fact_type.spec = _spec
    fact_type._fact_proxy.spec = _spec
    fact_type._fact_type_class._spec = _spec

    # Needs to be done because different definitions can share a 
    #  fact_type object
    if(hasattr(fact_type,'_clean_spec')): del fact_type._clean_spec

    out = [fact_type]
    if(return_proxy): out.append(fact_type._fact_proxy)
    if(return_type_class): out.append(fact_type._fact_type_class)
    return tuple(out) if len(out) > 1 else out[0]
    # return fact_ctor, fact_type


def define_facts(specs, #: list[dict[str,dict]],
                 context=None):
    '''Defines several facts at once.'''
    for name, spec in specs.items():
        define_fact(name,spec,context=context)

###### Base #####

base_fact_field_dict = {
    **cre_obj_field_dict,
}

base_fact_fields  = [(k,v) for k,v in base_fact_field_dict.items()]

BaseFact = _fact_from_fields("BaseFact", [])
register_global_default("Fact", BaseFact)

base_list_type = ListType(BaseFact)

# @lower_cast(Fact, CREObjType)
@lower_cast(Fact, BaseFact)
def upcast(context, builder, fromty, toty, val):
    return _obj_cast_codegen(context, builder, val, fromty, toty,incref=False)


@njit(i8(CREObjType), cache=True)
def fact_to_ptr(fact):
    return cast(fact, i8)

@njit(cache=True)
def fact_to_basefact(fact):
    return cast(fact, BaseFact)

@njit(cache=True)
def fact_to_ptr_incref(fact):
    return _ptr_from_struct_incref(fact)


###### Fact Casting #######
@generated_jit
def cast_fact(typ, val):
    '''Casts a fact to a new type of fact if possible'''
    from cre.context import cre_context
    context = cre_context()    
    inst_type = typ.instance_type

    #Check if the fact_type can be casted 
    if(inst_type._fact_name != "BaseFact" and val._fact_name != "BaseFact" and
       inst_type._fact_name not in context.children_of[val._fact_name] and 
       inst_type._fact_name not in context.parents_of[val._fact_name]
       
    ):
        error_message = f"Cannot cast fact of type '{val._fact_name}' to '{inst_type._fact_name}.'"
        #If it shouldn't be possible then throw an error
        def error(typ,val):
            raise TypeError(error_message)
        return error
    
    def impl(typ,val):
        return cast(val, inst_type)

    return impl


@njit(Tuple((u1,i8))(BaseFact), cache=True)
def get_inheritance_bytes_len_ptr(st):
    ptr = _struct_get_data_ptr(st) + st.chr_mbrs_infos_offset + \
             (st.num_chr_mbrs * _sizeof_type(member_info_type))
    num_inh_bytes = _load_ptr(u1, ptr)
    return num_inh_bytes, ptr+1


@njit(ListType(i8)(BaseFact), cache=True)
def get_inheritance_t_ids(st):
    nbytes, ptr = get_inheritance_bytes_len_ptr(st)
    t_ids = List.empty_list(i8)
    prev_byte = u1(-1)
    val = i8(0)
    prev_val = i8(0)
    for i in range(nbytes):
        byte = _load_ptr(u1,ptr+i)
        val = (val << 8) | byte
        if(byte != 0 and prev_byte == 0):
            t_ids.append(prev_val>>8)    
            val = i8(byte)
            
        prev_byte = byte
        prev_val = val
    t_ids.append(val)    
    return t_ids


# @njit()
# def fact_setitem(fact, index):





@generated_jit(cache=True,nopython=True)
@overload_method(Fact, "isa")
def isa(self, typ):
    typ = typ.instance_type


    if(typ is BaseFact or typ is CREObjType):
        def impl(self, typ):
            return True

    # If the type has defined it's own _isa then use that
    elif(hasattr(typ, '_isa')):
        _isa = typ._isa
        def impl(self, typ):
            return _isa(self)

    # Otherwise just check if the first t_id in inh_bytes is t_id of typ   
    #   This case handles cases like TupleFact that can only have one 
    #   level of inheritance.
    elif(hasattr(typ, "t_id")):
        t_id = typ.t_id
        inh_bytes = uint_to_inheritance_bytes(t_id)
        def impl(self, typ):
            l, p = get_inheritance_bytes_len_ptr(self)
            for i,b in enumerate(literal_unroll(inh_bytes)):
                f_b = _load_ptr(u1, p+i)
                if(b != f_b):
                    return False
            return True
    return impl

@generated_jit(cache=True,nopython=True)
@overload_method(Fact, "asa")
def asa(self, typ):
    # from numba.extending import SentryLiteralArgs
    # SentryLiteralArgs(['unsafe']).for_function(asa).bind(self, typ, unsafe) 

    _typ = typ.instance_type
    use_unsafe_cast = (_typ is CREObjType) or (_typ is BaseFact)

    # print("TRY CAST:", f'{str(self)} to {str(_typ)}')
    if(not use_unsafe_cast):
        error_message = f"Cannot cast fact of type '{str(self)}' to '{str(_typ)}.'"

        # from cre.context import cre_context
        # context = cre_context()    
        
        # print(context.children_of[self], _typ, _typ in context.children_of[self])
        # print(context.parents_of[self], _typ, _typ in context.parents_of[self])
        # #Check if the fact_type can be casted 
        # if(self in context.children_of and 
        #    _typ not in context.children_of[self] and 
        #    self in context.parents_of and
        #    _typ not in context.parents_of[self]
        # ):

        #     # error_message = f"Cannot cast fact of type '{val._fact_name}' to '{inst_type._fact_name}.'"
        #     #If it shouldn't be possible then throw an error
        #     def error(self, typ):
        #         raise TypeError(error_message)
        #     return error

        # fn1, fn2 = self._fact_name, _typ._fact_name
        _isa = typ.instance_type._isa
        def impl(self, typ):
            # print(self, _isa(self))
            if(not _isa(self)): raise TypeError(error_message)
            return cast(self, typ)
    else:
        def impl(self, typ):
            return cast(self, typ)
    return impl


# -------
# : Functions for dereferencing Facts

@njit(i8(deref_info_type,i8),inline='never',cache=True)
def deref_once(deref, inst_ptr):
    if(deref.type == u1(DEREF_TYPE_ATTR)):
        return _ptr_to_data_ptr(inst_ptr)
    else:
        return _list_base_from_ptr(inst_ptr)

@njit(i8(BaseFact, deref_info_type[::1]),cache=True)
def resolve_deref_data_ptr(fact, deref_infos):
    '''
    '''
    inst_ptr = cast(fact, i8)
    if(len(deref_infos) > 1):
        for k in range(len(deref_infos)-1):
            if(inst_ptr == 0): break;
            deref = deref_infos[k]
            data_ptr = deref_once(deref, inst_ptr)
            inst_ptr = _load_ptr(i8, data_ptr+deref.offset)

    if(inst_ptr != 0):
        deref = deref_infos[-1]
        data_ptr = deref_once(deref, inst_ptr)
        return data_ptr+deref.offset
    else:
        return 0


@generated_jit(cache=True)
def resolve_deref(self, deref_infos, out_type):
    def impl(self, deref_infos, out_type):
        data_ptr = resolve_deref_data_ptr(self, deref_infos)
        if(data_ptr == 0):
            raise AttributeError("Failed to dereference Fact.")
        return _load_ptr(out_type, data_ptr)
    return impl



#### Hashing ####

# @njit(cache=True)
# def fact_hash():





import numpy as np
from numba import types, jit,njit
from numba import void,b1,u1,u2,u4,u8,i1,i2,i4,i8,f4,f8,c8,c16
from numba.types import ListType, unicode_type
from cre.type_conv import str_to_float, str_to_int, int_to_str, float_to_str
from cre.var import Var


import timeit
N=1000
def time_ms(f):
    f() #warm start
    return " %0.6f ms" % (1000.0*(timeit.timeit(f, number=N)/float(N)))

def test_str_to_float():
    assert str_to_float("-12312.001345e-2")==float("-12312.001345e-2")
    assert str_to_float("12312.01e2")==float("12312.01e2")
    assert str_to_float("12312")==float("12312")
    assert str_to_float("1.3452e-12")==float("1.3452e-12")
    assert str_to_float("1.3452e12")==float("1.3452e12")
    # This one is harder to get exact, but it is close
    assert np.isclose(str_to_float("1.3452e25"),float("1.3452e25"))
    assert np.isclose(str_to_float("1.3452e-25"),float("1.3452e-25"))
    assert str_to_float("1.3452")==float("1.3452")
    assert str_to_float("-1000.9999")==float("-1000.9999")
    assert str_to_float("-0.009999")==float("-0.009999")
    assert str_to_float("0")==float("0")

def test_str_to_int():

    assert str_to_int("123") == int("123")
    assert str_to_int("-123") == int("-123")
    assert str_to_int("-1000023") == int("-1000023")
    # print(str_to_int("123.0") == int("123.0"))
    # print(str_to_int("-123.01") == int("-123.01"))
    # print(str_to_int("0.01") == int("0.01"))
    # print(str_to_int("-0.01") == int("-0.01"))

def to_str_close(x):
    # print(int_to_str(x), str(x))
    return np.isclose(float(float_to_str(x)), float(str(x)))


def test_to_str():
    # assert to_str_close(147)
    # assert to_str_close(-147) 
    assert to_str_close(146.78) 
    assert to_str_close(-146.78)
    assert to_str_close(0.78) 
    assert to_str_close(-0.78) 
    assert to_str_close(4.0)
    assert to_str_close(-4.0) 
    assert to_str_close(1.34e24) 
    assert to_str_close(-1.34e-24)
    assert to_str_close(2e24)
    assert to_str_close(2e-24)
    assert to_str_close(-2e24)
    assert to_str_close(-2e-24)
    assert to_str_close(9.9999999999e24)
    assert to_str_close(9.99999999999999e24)
    assert to_str_close(9.9999999999e-24)
    assert to_str_close(9.99999999999999e-24)
    assert to_str_close(-9.9999999999e24)
    assert to_str_close(-9.99999999999999e24)
    assert to_str_close(-9.9999999999e-24)
    assert to_str_close(-9.99999999999999e-24)
    assert to_str_close(-9.99999999999999e-200)
    assert to_str_close(9.99999999999999e+200)


@njit(cache=True)
def _str_test():
  return str(1.2)

@njit(cache=True)
def _float_test():
  return float("1.2")

def test_str_float_overloaded():
    assert _float_test() == 1.2
    assert _str_test() == '1.2'



if __name__ == "__main__":
    print(str_to_float("0"))
    # test_str_to_float()
    # test_to_str()
    # test_str_to_int()
    # test_ops()

    # @njit
    # def test_nb_float():
    #     str_to_float('-12312.0')

    # def test_py_float():
    #     float("-12312.0")   

    # @njit
    # def test_nb_int():
    #     str_to_int("-123")  

    # def test_py_int():
    #     int("-123") 


    # print("nb_float\t",time_ms(test_nb_float))
    # print("py_float\t",time_ms(test_py_float))
    # print("nb_int\t",time_ms(test_nb_int))
    # # print("nb_int_core\t",time_ms(test_nb_int_core))
    # print("py_int\t",time_ms(test_py_int))
    # print("float_to_str", time_ms(lambda : float_to_str(146.78)))




import numpy as np
import numba
from numba import njit, i8,i4,u1,u2,u4,u8, f8, f4, generated_jit, types, literal_unroll
from numba.typed import List, Dict
from numba.types import ListType, DictType, unicode_type, Tuple
from cre.fact import define_fact
from cre.structref import define_structref, define_structref_template
from numba.experimental import structref
from numba.experimental.structref import new, define_attributes
from numba.extending import lower_cast, overload, overload_method
from cre.memset import MemSet,MemSetType
from cre.utils import  cast, lower_setattr, _load_ptr, _ptr_from_struct_incref, decode_idrec, _listtype_sizeof_item, _obj_cast_codegen, DEREF_TYPE_ATTR, DEREF_TYPE_LIST, _memcpy_structref
from cre.vector import VectorType
from cre.transform.incr_processor import IncrProcessorType, ChangeEventType, incr_processor_fields, init_incr_processor
from cre.structref import CastFriendlyStructref, define_boxing
from cre.context import cre_context, CREContextDataType
from cre.fact import Fact, BaseFact, DeferredFactRefType
from cre.gval import gval as gval_type
from cre.var import VarType
from cre.vector import new_vector


def _get_base_type(fact_type):
    parent = None
    while(hasattr(fact_type,'parent_type')):
        parent = fact_type = fact_type.parent_type
    return parent if parent is not None else fact_type

def assert_mutual_base(fact_types, id_attr):
    if(len(fact_types) == 1): return fact_types[0]
    base = None
    for fact_type in fact_types:
        f_base = _get_base_type(fact_type)
        if(base is not None):
            assert f_base == base, "RelativeEncoder requires that all fact types inherit from a common base type."
        
        base = f_base

    assert id_attr in base.spec, "Base type must define an identifier attribute (e.g. like id='fact_instance_name')"
    return base



def get_relational_fact_attrs(fact_types):
    ''' Takes in a set of fact types and returns all (fact, attribute) pairs
        for "relational" attributes. 
    '''
    context = cre_context()

    rel_fact_attrs = {}
    for ft in fact_types:
        ft = ft.instance_type if (isinstance(ft, types.TypeRef)) else ft
        parents = context.parents_of.get(ft._fact_name,[])
        for attr, attr_spec in ft.filter_spec("relational").items():
            # relational = False
            attr_t = attr_spec['type']
            # attr_t = attr_t.instance_type if (isinstance(attr_t, (types.TypeRef,DeferredFactRefType))) else attr_t
            # if(isinstance(attr_t, Fact)): relational = True
            # if(isinstance(attr_t, types.ListType)):
            #     item_t = attr_t.item_type
            #     item_t = item_t.instance_type if (isinstance(item_t, (types.TypeRef,DeferredFactRefType))) else item_t
            #     if(isinstance(item_t, Fact)): relational = True
            #     attr_t = types.ListType(item_t)

            # relational = d.get("relational", relational)
                
            # if(relational):
            is_new = True
            for p in parents:
                if((p,attr) in rel_fact_attrs):
                    is_new = False
                    break
            if(is_new): rel_fact_attrs[(ft,attr)] = attr_t

    return tuple([(fact_type, attr_t, types.literal(attr)) for (fact_type, attr), attr_t in rel_fact_attrs.items()])


#### Relative Encoder ####
from cre.utils import deref_info_type, np_deref_info_type
np_ind_dist_deref = np.dtype([
    ("ind", np.int32),
    ("dist", np.float32),
    ("attr_id", np.uint32),
    ("item_ind", np.int32),
    # ('deref_info1', np_deref_info_type),
    # ('deref_info2', np_deref_info_type),
    ])
ind_dist_deref = numba.from_dtype(np_ind_dist_deref)

relative_encoder_fields = {
    **incr_processor_fields,
    "context_data" : CREContextDataType,
    "dist_matrix" : ind_dist_deref[:, ::1],
    "idrec_to_ind" : DictType(u8, i4),
    "visited_inds" : DictType(i8, u1),
    "facts" : ListType(BaseFact),
    "deref_info_matrix" : deref_info_type[:,::1],
    "lateral_w" : f8,
    "id_to_ind" : DictType(unicode_type, i4),
    "var_cache" : DictType(Tuple((i4,i4,i4)),VarType),
    "valid_t_ids" : VectorType,
    "keep_floating_facts" : types.boolean,
    "needs_rebuild" : types.boolean,
    "fact_types" : types.Any,
    "relational_attrs" : types.Any,
    "base_fact_type" : types.Any,
    "id_attr" : types.Any,
}

@structref.register
class RelativeEncoderTypeClass(CastFriendlyStructref):
    pass

GenericRelativeEncoderType = RelativeEncoderTypeClass([(k,v) for k,v in relative_encoder_fields.items()])

from numba.core.typing.typeof import typeof
def get_relative_encoder_type(fact_types, id_attr):
    relational_fact_attrs = get_relational_fact_attrs(fact_types)

    relational_fact_attrs_type = typeof(relational_fact_attrs)
    field_dict = {**relative_encoder_fields,
                 "fact_types" : Tuple(fact_types),
                 "relational_attrs" : relational_fact_attrs_type,
                 "base_fact_type" : types.TypeRef(assert_mutual_base(fact_types, id_attr)), 
                 "id_attr" : types.literal(id_attr),
                 }
                 
    re_type = RelativeEncoderTypeClass([(k,v) for k,v in field_dict.items()])
    re_type._fact_types = fact_types
    re_type._relational_attrs = relational_fact_attrs
    return re_type

class RelativeEncoder(structref.StructRefProxy):
    def __new__(cls, fact_types, in_memset=None, id_attr='id', context=None, keep_floating_facts=True):
        # Make new in_ms and out_ms if they are not provided.
        context = cre_context(context)
        fact_types=tuple(fact_types)
        if(in_memset is None): in_memset = MemSet(context);
        re_type = get_relative_encoder_type(fact_types,id_attr)
        self = RelativeEncoder_ctor(context.context_data, re_type, in_memset, keep_floating_facts)
        return self

    def encode_relative_to(self, ms, sources, source_vars):
        assert len(sources) == len(source_vars), f"{sources},{source_vars}"

        source_idrecs = np.empty(len(sources),dtype=np.uint64)
        for i, src in enumerate(sources): 
            source_idrecs[i] = src.idrec

        var_ptrs = np.empty(len(sources),dtype=np.int64)
        for i, v in enumerate(source_vars): 
            var_ptrs[i] = v.get_ptr()

        out_ms = encode_relative_to(self, ms, source_idrecs, var_ptrs)
        return out_ms

    @property
    def in_memset(self):
        return get_in_memset(self)

    def set_in_memset(self, in_memset):
        if(not check_same_in_memset(self,in_memset)):
            set_in_memset(self,in_memset)
            # RelativeEncoder_reinit(self)

    def update(self):
        RelativeEncoder_update(self)

define_boxing(RelativeEncoderTypeClass, RelativeEncoder)

@lower_cast(RelativeEncoderTypeClass, GenericRelativeEncoderType)
@lower_cast(RelativeEncoderTypeClass, IncrProcessorType)
def upcast(context, builder, fromty, toty, val):
    return _obj_cast_codegen(context, builder, val, fromty, toty, incref=False)


@njit(cache=True)
def new_dist_matrix(n, old_dist_matrix=None):
    dist_matrix = np.empty((n,n),dtype=ind_dist_deref)
    for i in range(n):
        for j in range(n):
            if(i == j):
                dist_matrix[i,j].ind = i4(-1)
                dist_matrix[i,j].dist = f4(0)
                dist_matrix[i,j].attr_id = u4(0)
                dist_matrix[i,j].item_ind = i4(-1)
            else:
                dist_matrix[i,j].ind = i4(-1)
                dist_matrix[i,j].dist = f4(np.inf)
                dist_matrix[i,j].attr_id = u4(0)
                dist_matrix[i,j].item_ind = i4(-1)
    if(old_dist_matrix is not None):
        o_n, o_m = old_dist_matrix.shape
        dist_matrix[:o_n, :o_m] = old_dist_matrix
    return dist_matrix



@generated_jit(cache=False, nopython=True)
def _fill_rel_derefs(deref_info_matrix, rel_derefs, start=0):
    if(len(rel_derefs) > 0):
        def impl(deref_info_matrix, rel_derefs, start=0):
            attr_id = start
            for d1 in literal_unroll(rel_derefs):
                # d1 = rel_derefs[i]
                deref_info_matrix[attr_id][0].type = d1[0]
                deref_info_matrix[attr_id][0].a_id = d1[1]
                deref_info_matrix[attr_id][0].t_id = d1[2]
                deref_info_matrix[attr_id][0].offset = d1[3]
                attr_id += 1
    else:
        def impl(deref_info_matrix, rel_derefs, start=0):
            pass
    return impl


@generated_jit(cache=False, nopython=True)
def _fill_list_derefs(deref_info_matrix, list_derefs, start=0):
    if(len(list_derefs) > 0):
        def impl(deref_info_matrix, list_derefs, start=0):
            attr_id = start
            for tup in literal_unroll(list_derefs):
                attr_t, d1, d2 = tup

                deref_info_matrix[attr_id][0].type = d1[0]
                deref_info_matrix[attr_id][0].a_id = d1[1]
                deref_info_matrix[attr_id][0].t_id = d1[2]
                deref_info_matrix[attr_id][0].offset = d1[3]
                deref_info_matrix[attr_id][1].type = d2[0]
                deref_info_matrix[attr_id][1].a_id = 0
                deref_info_matrix[attr_id][1].t_id = d2[2]
                item_size = _listtype_sizeof_item(attr_t)
                deref_info_matrix[attr_id][1].offset = item_size
                attr_id += 1
    else:
        def impl(deref_info_matrix, list_derefs, start=0):
            pass
    return impl

@generated_jit(cache=True, nopython=True)
def _build_valid_t_ids(self):
    context = cre_context()
    t_ids = set()
    for fact_type in self._fact_types:
        t_ids.add(context.get_t_id(_type=fact_type))
    val_t_ids = tuple(t_ids)
    max_t_id = int(max(val_t_ids))

    def impl(self):
        valid_t_ids = new_vector(max_t_id+1)
        for i in range(len(val_t_ids)):
            valid_t_ids.set_item_safe(val_t_ids[i], 1) 
        return valid_t_ids
    return impl

@generated_jit(cache=True, nopython=True)
def _build_deref_info_matrix(self):
    context = cre_context()
    re_type = self
    rel_derefs, list_derefs = [], []

    for base_t, attr_t, attr in re_type._relational_attrs:
        deref_info1 = base_t.get_attr_deref_info(attr.literal_value)
        if(not isinstance(attr_t,types.ListType)):
            rel_derefs.append(deref_info1.tolist())
        else:
            deref_info2 = np.zeros((),dtype=deref_info_type)
            deref_info2['type'] = DEREF_TYPE_LIST
            deref_info2['t_id'] = context.get_t_id(_type=attr_t.item_type)
            d1, d2 = deref_info1.tolist(), deref_info2.tolist()
            list_derefs.append((attr_t, deref_info1.tolist(), deref_info2.tolist()))        

    rel_derefs = tuple(rel_derefs)
    list_derefs = tuple(list_derefs)
    n_attrs = len(rel_derefs) + len(list_derefs)

    def impl(self):
        # Fill in the deref_info_matrix which we use to build Var instances 
        deref_info_matrix = np.zeros((n_attrs,2),dtype=deref_info_type)
        
        _fill_list_derefs(deref_info_matrix, list_derefs)
        _fill_rel_derefs(deref_info_matrix, rel_derefs,start=len(list_derefs))
        return deref_info_matrix

    return impl

var_cache_key_type = Tuple((i4,i4,i4))

@njit(cache=True)
def RelativeEncoder_reinit(self,size=32):
    self.dist_matrix = new_dist_matrix(size)
    self.idrec_to_ind = Dict.empty(u8,i4)
    self.visited_inds = Dict.empty(i8,u1)
    self.facts = List.empty_list(BaseFact)
    self.id_to_ind = Dict.empty(unicode_type,i4)
    self.var_cache = Dict.empty(var_cache_key_type,VarType)

# print("A")
@njit(cache=True)
def RelativeEncoder_ctor(context_data, re_type, in_memset, keep_floating_facts):
    # def impl(re_type, in_ms):
    st = new(re_type)
    init_incr_processor(st, in_memset)
    RelativeEncoder_reinit(st)
    st.context_data = context_data
    st.lateral_w = f8(1.0)
    st.deref_info_matrix = _build_deref_info_matrix(st)
    st.valid_t_ids = _build_valid_t_ids(st)
    st.needs_rebuild = False
    st.keep_floating_facts = keep_floating_facts
    return st

    
    return k+1
from cre.fact_intrinsics import fact_lower_getattr

# @njit(cache=True)
# def fill_adj_inds(self, fact, k, adj_inds, attr_id, item_ind):
#     ind = self.idrec_to_ind[fact.idrec]

#     adj_inds[k].ind = i4(ind)
#     adj_inds[k].dist = f4(self.lateral_w)
#     adj_inds[k].attr_id = u4(attr_id)
#     adj_inds[k].item_ind = i4(item_ind)

#     return k+1

@njit(cache=True)
def fill_adj_inds(self, fact, k, adj_inds, attr_id, item_ind):
    ind = self.idrec_to_ind[fact.idrec]
    adj_inds[k].ind = i4(ind)
    adj_inds[k].dist = f4(self.lateral_w - (1.0/(2<<(attr_id+1)*3) if attr_id < 7 else 0.0))
    adj_inds[k].attr_id = u4(attr_id)
    adj_inds[k].item_ind = i4(item_ind)
    return k+1

@njit(cache=True)
def fill_adj_list_inds(self, lst, k, adj_inds, attr_id):
    for i, item in enumerate(lst):
        fact = cast(item, self.base_fact_type)
        # k = fill_adj_inds(self, base_item, k, adj_inds, attr_id, i4(i))
        ind = self.idrec_to_ind[fact.idrec]
        adj_inds[k].ind = i4(ind)
        adj_inds[k].dist = f4(self.lateral_w - (1.0/(2<<(attr_id+1)*3) if attr_id < 7 else 0.0))
        adj_inds[k].attr_id = u4(attr_id)
        adj_inds[k].item_ind = i4(i)
        k += 1
    return k


@generated_jit(cache=True,nopython=True)
def _next_list_adj(self, fact):
    list_attrs = []
    n_rel = 0
    for base_t, attr_t, attr in self._relational_attrs:
        if(isinstance(attr_t,types.ListType)):
            list_attrs.append((attr, base_t))
        else:
            n_rel += 1
    list_attrs = tuple(list_attrs)

    if(len(list_attrs) == 0): 
        return lambda self, fact : (np.empty(n_rel, dtype=ind_dist_deref), 0, 0)
    
    def impl(self, fact):
        # Make a buffer to fill adj_inds
        n_ladj = 0
        for ltup in literal_unroll(list_attrs):
            attr, base_t = ltup
            if(fact.isa(base_t)):
                typed_fact = cast(fact, base_t)
                n_ladj += len(fact_lower_getattr(typed_fact, attr))
        adj_inds = np.empty(n_rel+n_ladj, dtype=ind_dist_deref)
        # adj_facts = List.empty_list(self.base_fact_type)
        attr_id, k = 0, 0

        # Fill in list relative attributes
        for ltup in literal_unroll(list_attrs):
            attr, base_t = ltup
            if(fact.isa(base_t)):
                typed_fact = cast(fact, base_t)

                # Member is a list so go through it's items
                lst = fact_lower_getattr(typed_fact, attr)
                k = fill_adj_list_inds(self, lst, k, adj_inds, attr_id)
                # for i, item in enumerate(member):
                #     base_item = cast(item, self.base_fact_type)
                #     k = fill_adj_inds(self, base_item, k, adj_inds, attr_id, i4(i))
            attr_id += 1
        return adj_inds, attr_id, k
    return impl

@generated_jit(cache=True,nopython=True)
def _next_rel_adj(self, fact, adj_inds, attr_id, k):
    rel_attrs = []
    n_rel = 0
    for base_t, attr_t, attr in self._relational_attrs:
        if(not isinstance(attr_t,types.ListType)):
            rel_attrs.append((attr, base_t))
    rel_attrs = tuple(rel_attrs)
    # print("<<<<", rel_attrs)

    if(len(rel_attrs) == 0): 
        return lambda self, fact, adj_inds, attr_id, k : (adj_inds, attr_id, k)

    def impl(self, fact, adj_inds, attr_id, k):
        # Fill in list relative attributes
        for tup in literal_unroll(rel_attrs):
            attr, base_t = tup
            if(fact.isa(base_t)):
                typed_fact = cast(fact, base_t)
                member = fact_lower_getattr(typed_fact, attr)

                if(member is not None):
                    base_member = cast(member, self.base_fact_type)
                    k = fill_adj_inds(self, base_member, k, adj_inds, attr_id, -1)
            attr_id += 1
        return adj_inds, attr_id, k
    return impl
        
@njit(cache=True)
def next_adjacent(self, fact):
    adj_inds, attr_id, k = _next_list_adj(self, fact)
    adj_inds, attr_id, k = _next_rel_adj(self, fact, adj_inds, attr_id, k)
    return adj_inds[:k]


@njit(cache=True)
def update_relative_to(self , sources):
    dist_matrix = self.dist_matrix
    frontier_inds = Dict.empty(i8,u1)
    next_frontier_inds = Dict.empty(i8,u1)

    for src in sources:
        s_ind = self.idrec_to_ind[src.idrec]
        frontier_inds[i8(s_ind)] = u1(1)

    while(len(frontier_inds) > 0):
        for b_ind in frontier_inds:
            b = self.facts[b_ind]
            for c_ind_dist in next_adjacent(self,b):

                c_ind = i8(c_ind_dist.ind)

                # Skip if does not exist
                if(c_ind == -1): continue
                    
                # If the dist_matrix doesn't have b->c then fill it in
                if(dist_matrix[b_ind, c_ind].ind == -1):
                    #Assign new heading and dist
                    dist_matrix[b_ind, c_ind].ind = c_ind
                    dist_matrix[b_ind, c_ind].dist = c_ind_dist.dist 
                    dist_matrix[b_ind, c_ind].attr_id = c_ind_dist.attr_id
                    dist_matrix[b_ind, c_ind].item_ind = c_ind_dist.item_ind

                    next_frontier_inds[c_ind] = u1(1)

                    # print("CONNECT:", f'{b_ind}', "->", f'{c_ind}', c_ind_dist.dist)
                    # print("CONNECT:", f'{b.id}', "->", f'{c.id}', c_ind_w_pair.dist)

                for a_ind in self.visited_inds:
                    if( a_ind == b_ind or a_ind == c_ind): continue

                    new_dist = dist_matrix[a_ind, b_ind].dist  + dist_matrix[b_ind, c_ind].dist 
                    if(new_dist < dist_matrix[a_ind, c_ind].dist):
                        # print("CONNECT:", f'{self.comps[a_ind].id}', "->", f'{c.id},', "via", f'{b.id}', ":", new_dist, "<", dist_matrix[a_ind, c_ind].dist)
                        dist_matrix[a_ind, c_ind].ind = dist_matrix[a_ind, b_ind].ind 
                        dist_matrix[a_ind, c_ind].dist = new_dist 
                        dist_matrix[a_ind, c_ind].attr_id = dist_matrix[a_ind, b_ind].attr_id
                        dist_matrix[a_ind, c_ind].item_ind = dist_matrix[a_ind, b_ind].item_ind
                        next_frontier_inds[c_ind] = u1(1)
                
            self.visited_inds[b_ind] = u1(1)
        frontier_inds = next_frontier_inds
        next_frontier_inds = Dict.empty(i8,u1)

    # print("B")
    # print(dist_matrix[:len(self.idrec_to_ind),:len(self.idrec_to_ind)])


@generated_jit(cache=True, nopython=True)
@overload_method(RelativeEncoderTypeClass,'get_changes')
def incr_pr_accumulate_change_events(self, end=-1, exhaust_changes=True):
    def impl(self, end=-1, exhaust_changes=True):
        incr_pr = cast(self, IncrProcessorType)
        return incr_pr.get_changes(end=end, exhaust_changes=exhaust_changes)
    return impl


@generated_jit(cache=True,nopython=True)
def _check_needs_rebuild(self, change_events):
    context = cre_context()
    # Make base_t, a_id pairs for relative attributes
    base_a_id_pairs = []
    for base_t, attr_t, attr in self._relational_attrs:
        a_id = base_t.get_attr_a_id(attr.literal_value)
        base_a_id_pairs.append((base_t,a_id))
    base_a_id_pairs = tuple(base_a_id_pairs)

    def impl(self, change_events):
        # Need rebuild if there were any retractions or modifications to
        #  'relational' attributes.
        need_rebuild = False
        for ce in change_events:
            # Ensure that the t_ids belong to provided fact_types.
            if(ce.t_id <= len(self.valid_t_ids) and self.valid_t_ids[i8(ce.t_id)]):
                if(ce.was_retracted): 
                    need_rebuild = True; break;
                if(ce.was_modified):
                    fact = self.in_memset.get_fact(ce.idrec, self.base_fact_type)
                    for tup in literal_unroll(base_a_id_pairs):
                        base_t, a_id = tup
                        if(a_id in ce.a_ids and fact.isa(base_t)):
                            need_rebuild = True; break;
            else:
                raise ValueError(
"Relative Encoder in_memset should only contain facts of given fact_types. \
Ensure that RelativeEncoder is not initialized with a flattened MemSet."
                    )
            
        return need_rebuild
            
    return impl


@njit(cache=True)
def _insert_fact(self,fact):
    ind = i4(len(self.idrec_to_ind))
    self.idrec_to_ind[fact.idrec] = ind
    self.id_to_ind[fact_lower_getattr(fact, self.id_attr)] = ind
    self.facts.append(fact)


@njit(cache=True)
def _ensure_dist_matrix_size(self):
    if(len(self.dist_matrix) < len(self.facts)):
        self.dist_matrix = new_dist_matrix(len(self.facts)*2, self.dist_matrix)
    


@njit(cache=True)
def RelativeEncoder_update(self):
    change_events = self.get_changes()
    
    # Need rebuild if there were any retractions or modifications to
    #  'relative' attributes. TODO: rebuilds could probably triggered
    #   more conservatively.
    if(not self.needs_rebuild):
        self.needs_rebuild = _check_needs_rebuild(self,change_events)
    # print("<<", "need_rebuild", need_rebuild)
    if(self.needs_rebuild):
        # If need_rebuild start bellman-ford shortest path from scratch
        RelativeEncoder_reinit(self)
        sources = self.in_memset.get_facts(self.base_fact_type)
        for fact in sources:
            _insert_fact(self, fact)
        _ensure_dist_matrix_size(self)
        update_relative_to(self, sources)    
    else:
        # Otherwise we can do an incremental update
        # TODO: For some reason slower on second run. Might be in here
        #  although could also be a cache locality thing.
        sources = List.empty_list(self.base_fact_type)
        for ce in change_events:
            if(ce.was_declared or ce.was_modified):
                if(ce.was_declared):
                    fact = self.in_memset.get_fact(ce.idrec, self.base_fact_type)
                    _insert_fact(self,fact)
                    sources.append(fact)
        _ensure_dist_matrix_size(self)
        update_relative_to(self, sources)    

    self.needs_rebuild = False

@njit(cache=True)
def _closest_source(self, ind, s_inds):
    '''Finds the closest source index among 's_inds' to the fact 
         at index 'ind'. Distance is src->fact, not other way around.'''
    min_i = -1
    min_s_ind = -1
    min_dist = np.inf
    for i, s_ind in enumerate(s_inds):
        dist = self.dist_matrix[i8(s_ind)][i8(ind)].dist
        if(dist < min_dist): 
            min_dist = dist
            min_s_ind = s_ind
            min_i = i

    return min_s_ind, min_i

@njit(cache=True)
def _new_rel_var(source_var, head_t_id, deref_infos):
    ''' A constructor for a new Var instance with base source_var'''
    # source_var = source_vars[s_i]    
    # new_var = _memcpy_structref(source_var)
    new_var = new(VarType)
    new_var.idrec = source_var.idrec
    # new_var.alias = source_var.alias
    new_var.base_t_id = source_var.base_t_id
    new_var.head_t_id = head_t_id
    new_var.base_ptr = source_var.base_ptr
    new_var.base_ptr_ref = _ptr_from_struct_incref(source_var)
    new_var.is_not = u1(0)
    new_var.conj_ptr = i8(0)
    new_var.deref_infos = deref_infos
    new_var.hash_val = u8(0)
    return new_var

@njit(cache=True)
def _make_rel_var(self, f_ind, s_ind, s_var, extra_derefs=None):
    '''Builds a Var instance with a dereference chain 
         id==id_str. Encodes relative to the closest source. '''

    # Follow the shortest path encoded in the Bellman-Ford dist_matrix 
    #  to build the deref_info sequence for the var
    deref_info_buffer = np.empty((2*len(self.facts),),dtype=deref_info_type)
    k = 0; ind = s_ind;
    while(True):
        dm_entry = self.dist_matrix[i8(ind)][i8(f_ind)]
        ind = dm_entry.ind

        if(ind == i4(-1)): break

        deref_infos = self.deref_info_matrix[dm_entry.attr_id]
        deref_info_buffer[k:k+1] = deref_infos[0:1]
        k += 1
        if(deref_infos[1].type == DEREF_TYPE_LIST):
            deref_info_buffer[k].type = DEREF_TYPE_LIST
            deref_info_buffer[k].a_id = dm_entry.item_ind
            deref_info_buffer[k].t_id = decode_idrec(self.facts[ind].idrec)[0]#deref_infos[1].t_id
            deref_info_buffer[k].offset = dm_entry.item_ind*deref_infos[1].offset
            k += 1
        else:
            deref_info_buffer[k-1].t_id = decode_idrec(self.facts[ind].idrec)[0]

    # Add any extra deref_infos. e.g.'.value' at the end of a deref chain.
    if(extra_derefs is not None):
        for i in range(len(extra_derefs)):
            deref_info_buffer[k:k+1] = extra_derefs[i:i+1]
            k += 1

    # Make sure that the last t_id is the actual t_id of the fact
    #  so that we are certain we can dereference the final value.
    if(k >= 2):
        # print("&&", deref_info_buffer[k-2])
        # p = k-2 if deref_info_buffer[k-2].type == DEREF_TYPE_ATTR else k-3
        deref_info_buffer[k-2].t_id = decode_idrec(self.facts[f_ind].idrec)[0]
        # print("&&", deref_info_buffer[k-2].t_id)
    
    # Make the Var
    head_t_id = deref_info_buffer[k-1].t_id if(k > 0) else s_var.base_t_id
    # print(deref_info_buffer[:k].copy(), s_var, s_var.base_t_id, head_t_id)
    r_var =  _new_rel_var(s_var, head_t_id, deref_info_buffer[:k].copy())
    # print(r_var.base_t_id)
    return  r_var


@njit(cache=True)
def f_ind_from_id(self, id_str):
    if(id_str not in self.id_to_ind): 
        print(f"No fact with {self.id_attr}={id_str}")
        raise ValueError(f"Tried to re-encode fact with an id unknown to the RelativeEncoder.")
    f_ind = self.id_to_ind[id_str]
    return f_ind

@njit(cache=True,locals={"tup": var_cache_key_type})
def get_rel_var(self, f_ind, source_vars, s_inds, extra_derefs=None):
    '''Builds a Var using the relative encoder for the fact instance with
         id==id_str. Var has dereference chain that starts at the closest source. '''
    s_ind, s_i = _closest_source(self, f_ind, s_inds)

    if(s_ind == -1):
        return None

    s_var = source_vars[s_i]

    offset, cachable = i4(-1), True
    if(extra_derefs is not None):
        if(len(extra_derefs) > 1):
            cachable = False
        elif(len(extra_derefs) == 1):
            offset = extra_derefs[0].offset

    if(cachable):
        tup = (i4(s_ind),i4(f_ind), offset)
        if(tup not in self.var_cache):
            rel_var = _make_rel_var(self, f_ind, s_ind, s_var, extra_derefs)
            self.var_cache[tup] = rel_var
            return rel_var
        else:
            return self.var_cache[tup]
    else:
        return _make_rel_var(self, f_ind, s_ind, s_var, extra_derefs)

    # return _make_rel_var(self, f_ind, s_ind, source_vars[s_i])
from cre.core import T_ID_TUPLE_FACT, T_ID_VAR
from cre.obj import copy_cre_obj, _iter_mbr_infos, cre_obj_set_item, CREObjType, PRIMITIVE_MBR_ID, OBJECT_MBR_ID
from cre.tuple_fact import TupleFact


@njit(cache=True, locals={'rel_var' : VarType})
def rel_encode_tf(self, tf, source_vars, s_inds):
    tf_copy = copy_cre_obj(tf)
    tf_cre_obj = cast(tf, CREObjType)
    for i, (t_id, m_id, data_ptr) in enumerate(_iter_mbr_infos(tf_cre_obj)):
        if(t_id == T_ID_VAR):
            v = cast(_load_ptr(i8,data_ptr), VarType)
            f_ind = f_ind_from_id(self, v.alias)
            _rel_var = get_rel_var(self, f_ind, source_vars, s_inds, v.deref_infos)
            if(_rel_var is None): 
                _rel_var = v
                is_floating = True
            rel_var = _rel_var
            cre_obj_set_item(tf_copy, i, rel_var)
            # cre_obj_set_member_info(tf_copy, i, (T_ID_VAR, OBJECT_MBR_ID))

        elif(m_id != PRIMITIVE_MBR_ID):

            # Fact case 
            fact = cast(_load_ptr(i8,data_ptr), CREObjType)
            if(fact.idrec in self.idrec_to_ind):
                f_ind = self.idrec_to_ind[fact.idrec]
                _rel_var = get_rel_var(self, f_ind, source_vars, s_inds)
                if(_rel_var is not None):
                    rel_var = _rel_var
                    cre_obj_set_item(tf_copy, i, rel_var)
                    # cre_obj_set_member_info(tf_copy, i, (T_ID_VAR, OBJECT_MBR_ID))




    return tf_copy


@njit(cache=True, locals={"source_idrecs" : u8[::1]})
def encode_relative_to(self, gval_ms, source_idrecs, source_var_ptrs):
    # Ensure that it is up-to-date.
    RelativeEncoder_update(self)
    # Get the inds associated with the source_idrecs
    source_vars = List.empty_list(VarType)
    for i, var_ptr in enumerate(source_var_ptrs):
        source_vars.append(cast(var_ptr, VarType))

    s_inds = np.empty((len(source_idrecs,)),dtype=np.int32)
    for i, s_idrec in enumerate(source_idrecs):
        if(s_idrec not in self.idrec_to_ind):
            raise ValueError("Source fact not declared to input MemSet.")
        s_inds[i] = self.idrec_to_ind[s_idrec]

    rel_ms = MemSet(self.context_data)

    for fact in gval_ms.get_facts(gval_type):
        fact_copy = copy_cre_obj(fact)

        head = fact.head
        head_t_id, _, _ = decode_idrec(head.idrec)
        is_floating = False
        # For TupleFacts replace each Var member with a relatively encoded one.
        if(head_t_id == T_ID_TUPLE_FACT):
            tf = cast(head, TupleFact)
            new_tf = rel_encode_tf(self, tf, source_vars, s_inds)
            # print(tf, "->", new_tf)
            fact_copy.head = cast(new_tf, CREObjType)

        elif(head_t_id == T_ID_VAR):
            v = cast(head, VarType)
            f_ind = f_ind_from_id(self, v.alias)
            rel_var = get_rel_var(self, f_ind, source_vars, s_inds, v.deref_infos)
            if(rel_var is None): 
                rel_var = v
                is_floating = True
            fact_copy.head = cast(rel_var, CREObjType)
        


        if(self.keep_floating_facts or not is_floating):
            rel_ms.declare(fact_copy)
    return rel_ms
            

@njit(types.boolean(GenericRelativeEncoderType, MemSetType),cache=True)
def check_same_in_memset(self,in_memset):
    return cast(self.in_memset, i8) == cast(in_memset, i8)

@njit(types.void(GenericRelativeEncoderType, MemSetType), cache=True)
def set_in_memset(self, in_memset):
    self.in_memset = in_memset
    self.needs_rebuild = True
    # update_fact_vecs(self)

@njit(MemSetType(GenericRelativeEncoderType),cache=True)
def get_in_memset(self):
    return self.in_memset



from setuptools import setup, find_packages, Extension
import os, sys, sysconfig
import re

# Try to get the version from 
VERSIONFILE="cre/version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    __version__ = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

# TODO: Remove once version dependancy issues in caching are fixed
__version__ = "0.4.3"


# Read requirements.txt for requirements
with open('requirements.txt') as f: 
    requirements = f.readlines() 

  
long_description = 'CRE is a high-performance forward-chaining rule engine for Python built with the numba just-in-time compiler toolset.' 

dev_requirements = [
    "pytest",
    "pytest-benchmark"
]


def ensure_numba():
    from importlib import import_module
    print("Start")
    numba = None
    # while(numba is None):
    try:
        print("try")
        numba = import_module("numba")
    except ModuleNotFoundError:
        print("NO Numba")
        try:
            import pip
            for numba_req in requirements:
                if("numba" in numba_req):
                    break
            print("numba_req", numba_req)
            if("numba" not in numba_req):
                return
            thing = pip.main(['install', numba_req])  
        except:
            pass
    try:
        numba = import_module("numba")
    except ModuleNotFoundError:
        raise ModuleNotFoundError("Failed to install module 'numba' automatically.")
    return numba


def get_ext_modules():
    # Check that python development headers are installed
    python_headers_path = sysconfig.get_config_vars()['INCLUDEPY']
    if not os.path.exists(python_headers_path):
        major = sys.version_info.major
        minor = sys.version_info.minor
        raise ModuleNotFoundError(
    f"Ensure python headers installed: `sudo apt-get install libpython{major}.{minor}-dev`"
        )

    include_dirs = [sysconfig.get_path('include')]

    # Numba is required ahead of time to compile the cfunc extension module.
    #  But if installing from PyPi then we'll get it precompiled.
    try:
        numba = ensure_numba()
        numba_path = numba.extending.include_path()
        include_dirs = [numba_path, sysconfig.get_path('include')]
    except Exception as e:
        print("Getting numba path raised:")
        print(e)
        pass

    print("INCLUDE DIRS:", include_dirs)

    cre_c_funcs = Extension(
        name='cre_cfuncs', 
        sources=['cre/cfuncs/cre_cfuncs.c'],
        include_dirs=include_dirs
    )
    return [cre_c_funcs]

  
setup( 
        name ='cre', 
        version = __version__, 
        author ='Daniel Weitekamp', 
        author_email ='dannyweitekamp@gmail.com', 
        url ='https://github.com/DannyWeitekamp/Cognitive-Rule-Engine', 
        description ='A rule engine for Python powered by numba.', 
        long_description = long_description, 
        long_description_content_type ="text/markdown", 
        license ='MIT', 
        packages = find_packages(), 
        include_package_data=True, # Prevents files being omitting from wheel. 

        entry_points={
            "console_scripts": [
                "cre = console.cre_exec:main"
            ]
        },
        ext_modules = get_ext_modules(),


        classifiers =[ 
            "Programming Language :: Python :: 3", 
            "License :: OSI Approved :: MIT License", 
            "Operating System :: OS Independent", 
            "Development Status :: 2 - Pre-Alpha"
        ], 
        keywords = ['rule engine', 'expert system', 'production rules'], 
        install_requires = requirements, 
        setup_requires = ['numba'],
        extras_require={
            'dev' : dev_requirements
        }
) 



import numpy as np
import numba
from numba import njit, i8,i4,u1, f8, f4, generated_jit
from numba.typed import List, Dict
from numba.types import ListType, DictType
from cre.structref import define_structref, define_structref_template
from numba.experimental.structref import new, define_attributes
from numba.extending import lower_cast, overload_method
from cre.memset import MemSet,MemSetType
from cre.utils import _obj_cast_codegen
from cre.vector import VectorType
from cre.change_event import ChangeEventType, accumulate_change_events


########### IncrProcessor ##########

from numba import u1,u2,u8, types
from numba.types import Tuple, Set
from cre.fact import BaseFact
from cre.utils import decode_idrec, encode_idrec

incr_processor_fields = {
    "in_memset" : MemSetType,
    "change_queue_head" : i8,
}

IncrProcessor, IncrProcessorType, IncrProcessorTypeTemplate = define_structref("IncrProcessor", incr_processor_fields, define_constructor=True, return_type_class=True) 
IncrProcessorTypeTemplate.__str__ = lambda x : "cre.IncrProcessor"    


# print(IncrProcessor)
# print(IncrProcessorType)
# raise ValueError()

@njit(IncrProcessorType(IncrProcessorType, MemSetType),cache=True)
def init_incr_processor(st, ms):
    st.in_memset = ms
    st.change_queue_head = 0
    return st

# @njit(IncrProcessorType(IncrProcessorType, MemSetType),cache=True)
# def incr_processor_ctor(st, mem):
#     st = new(IncrProcessorType)
#     st.in_mem = mem
#     st.change_queue_head = 0
#     return st






@generated_jit(cache=True)
@overload_method(IncrProcessorTypeTemplate,'get_changes')
def incr_pr_accumulate_change_events(incr_pr, end=-1, exhaust_changes=True):
    def impl(incr_pr, end=-1, exhaust_changes=True):
        cq = incr_pr.in_memset.change_queue
        start = incr_pr.change_queue_head
        if(end == -1): end = cq.head
        change_events = accumulate_change_events(cq, start, end)
        if(exhaust_changes): incr_pr.change_queue_head = end
        return change_events
    return impl



def get_changes(incr_pr, end=-1, exhaust_changes=True):
    return incr_pr_accumulate_change_events(incr_pr, end, exhaust_changes)

IncrProcessor.get_changes = get_changes
